import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Collections;

/**
 * This class consists of methods used to sort lists of integer values using
 * different implementations of sorting algorithms. Some of these
 * implementations were obtained from the course textbook (Lewis & Chase,
 * <em>Java Software Structures: Designing and Using Data Structures</em>, 4th
 * Edition) and modified by Dr. Vargas-Perez.
 * 
 * @author Sandino Vargas-Perez
 * @version May 8, 2020
 *
 */
public class Sorters {

	/**
	 * Sorts the specified list of integer values using the Selection Sort
	 * algorithm. This implementation's time complexity is
	 * <em>O(n<sup>2</sup>)</em>.
	 * 
	 * @param list the ArrayList of integer values to be sorted
	 */
	public static void selectionSort(ArrayList<Integer> list) {
		int min;

		for (int index = 0; index < list.size() - 1; index++) {
			min = index;
			for (int scan = index + 1; scan < list.size(); scan++) {
				if (list.get(scan) < list.get(min))
					min = scan;
			}
			swap(list, min, index);
		}
	}

	/**
	 * Sorts the specified list of integer values using the Insertion Sort
	 * algorithm. This implementation's time complexity is
	 * <em>O(n<sup>2</sup>)</em>, but it can be <em>O(n)</em> if the list is sorted.
	 * 
	 * @param list the ArrayList of integer values to be sorted
	 */
	public static void insertionSort(ArrayList<Integer> list) {
		for (int index = 1; index < list.size(); index++) {
			int key = list.get(index);
			int position = index;

			// Shift larger values to the right
			while (position > 0 && list.get(position - 1) > key) {
				list.set(position, list.get(position - 1));
				position--;
			}

			list.set(position, key);
		}
	}

	/**
	 * Sorts the specified list of integer values using the Bubble Sort algorithm.
	 * This implementation's time complexity is <em>O(n<sup>2</sup>)</em>.
	 * 
	 * @param list the ArrayList of integer values to be sorted
	 */
	public static void bubbleSort1(ArrayList<Integer> list) {
		for (int position = list.size() - 1; position >= 0; position--) {
			for (int scan = 0; scan <= position - 1; scan++) {
				if (list.get(scan) > list.get(scan + 1)) {
					swap(list, scan, scan + 1);
				}
			}
		}
	}

	/**
	 * Sorts the specified list of integer values using the Bubble Sort algorithm.
	 * This implementation's time complexity is <em>O(n<sup>2</sup>)</em> but it can
	 * be <em>O(n)</em> if the list is sorted.
	 * 
	 * @param list the ArrayList of integer values to be sorted
	 */
	public static void bubbleSort2(ArrayList<Integer> list) {
		int position = list.size() - 1;
		boolean swapped = true;

		while (swapped) {
			swapped = false;
			for (int scan = 0; scan <= position - 1; scan++) {
				if (list.get(scan) > list.get(scan + 1)) {
					swap(list, scan, scan + 1);
					swapped = true;
				}
			}
			position--;
		}
	}

	/**
	 * Sorts the specified list of integer values using the Quick Sort algorithm.
	 * This implementation's time complexity is <em>O(n log n)</em>, but can be
	 * <em>O(n<sup>2</sup>)</em> if the chosen pivot is always the smallest or the
	 * biggest.
	 * 
	 * @param list the ArrayList of integer values to be sorted
	 */
	public static void quickSort(ArrayList<Integer> list) {
		quickSort(list, 0, list.size() - 1);
	}

	/**
	 * Recursively sorts a range of integers in the specified list using the Quick
	 * Sort algorithm. This method is invoked by
	 * <code>quickSort(ArrayList&lt;Integer&gt; list)</code>.
	 * 
	 * @param list the ArrayList of integer values to be sorted
	 * @param min  the index of the first element
	 * @param max  the index of the last element
	 */
	private static void quickSort(ArrayList<Integer> list, int min, int max) {
		if (min < max) {
			// Create partition with an element in the middle
			int indexOfPivot = partition(list, min, max);
			// Sort the left partition (values smaller than pivot)
			quickSort(list, min, indexOfPivot - 1);
			// Sort the right partition (values greater than pivot)
			quickSort(list, indexOfPivot + 1, max);
		}
	}

	/**
	 * Sorts the specified list of integer values using the Merge Sort algorithm.
	 * This implementation's time complexity is <em>O(n log n)</em>.
	 * 
	 * @param list the ArrayList of integer values to be sorted
	 */
	public static void mergeSort(ArrayList<Integer> list) {
		mergeSort(list, 0, list.size() - 1);
	}

	/**
	 * Recursively sorts a range of integers in the specified list using the Merge
	 * Sort algorithm. This method is invoked by
	 * <code>mergeSort(ArrayList&lt;Integer&gt; list)</code>.
	 * 
	 * @param list the ArrayList of integer values to be sorted
	 * @param min  the index of the first element
	 * @param max  the index of the last element
	 */
	private static void mergeSort(ArrayList<Integer> list, int min, int max) {
		if (min < max) {
			int mid = (min + max) / 2;
			mergeSort(list, min, mid);
			mergeSort(list, mid + 1, max);
			merge(list, min, mid, max);
		}
	}

	/**
	 * Recursively sorts a range of integers in the specified list using the Radix
	 * Sort algorithm. This implementation's time complexity is <em>O(C*n)</em>,
	 * where <em>C</em> is a constant representing the maximum number of digits in
	 * the elements of the list, or <em>O(n)</em>. If <em>C</em> is closer to
	 * <em>n</em>, then Radix Sort will have a time complexity of
	 * <em>O(n<sup>2</sup>)</em>.
	 * 
	 * 
	 * @param list the ArrayList of integer values to be sorted
	 */
	public static void radixSort(ArrayList<Integer> list) {
		// Represents the current digit: 1st, 10th, 100th..
		int ith = 1;
		// Max number of digits in the numbers of the list
		int c = String.valueOf(getMax(list)).length();

		// Create the digit queues for 0 to 9
		ArrayList<LinkedList<Integer>> digitQueue = new ArrayList<LinkedList<Integer>>();
		for (int i = 0; i < 10; i++) {
			digitQueue.add(new LinkedList<Integer>());
		}

		while (c > 0) {
			for (Integer number : list) {
				// Gets the corresponding digit from the current number
				int digit = (number % (ith * 10)) / ith;
				// Enqueue the numbers in the corresponding digit queue
				digitQueue.get(digit).add(number);
			}

			// Gather the numbers back into the list
			//list.clear();
			int j = 0;
			for (int digit = 0; digit < 10; digit++) {
				while (!digitQueue.get(digit).isEmpty()) {
					// Dequeue the numbers from the corresponding digit queue
					// and enqueue them in the list
					list.set(j,digitQueue.get(digit).remove());
					j++;
				}
			}

			// Gets the next digit.
			ith *= 10;
			c--;
		}
	}

	/**
	 * Sorts the specified list of integer values using the Bogo Sort algorithm,
	 * a.k.a. Stupid Sort. This implementation's time complexity can be
	 * <em>O(n!)</em>, but it will generally tend to be <em>O(&infin;)</em>, which
	 * means this algorithm might run "forever".
	 * 
	 * @param list the ArrayList of integer values to be sorted
	 */
	public static void bogoSort(ArrayList<Integer> list) {
		while (!isSorted(list)) {
			Collections.shuffle(list);
		}
	}
	/**
	 * 
	 * @param list
	 */
	public static void countingSort(ArrayList<Integer> list) 
    { 
        int n = list.size(); 
		int k = getMax(list) + 1;
        // The output integer array that will have sorted list 
        int output[] = new int[n]; 
  
        // Create a count array to store count of individual 
        // numbers and initialize count array as 0 
        int count[] = new int[k]; 

        // store count of each character 
        for (int i = 0; i<n; ++i) 
            ++count[list.get(i)]; 
  
        // Change count[i] so that count[i] now contains actual 
        // position of this character in output array 
        for (int i=1; i<k; ++i) 
            count[i] += count[i-1]; 
  
        // Build the output character array 
        // To make it stable we are operating in reverse order. 
        for (int i = n-1; i>=0; i--) 
        { 
            output[count[list.get(i)]-1] = list.get(i); 
            --count[list.get(i)]; 
        } 
        // Copy the output array to arr, so that arr now 
        // contains sorted characters 
        for (int i = 0; i<n; ++i) 
        	list.set(i, output[i]); 
    } 

	/**
	 * Helper method for swapping two values in a list.
	 * 
	 * @param list   the ArrayList of integer elements
	 * @param index1 index of the first element in the list to be swapped
	 * @param index2 index of the second element in the list to be swapped
	 */
	private static void swap(ArrayList<Integer> list, int index1, int index2) {
		int temp = list.get(index1);
		list.set(index1, list.get(index2));
		list.set(index2, temp);
	}

	/**
	 * Helper method to verify if a list is sorted.
	 * 
	 * @param list the ArrayList of integer elements
	 * @return true if list is sorted, false otherwise
	 */
	private static boolean isSorted(ArrayList<Integer> list) {
		for (int i = 0; i < list.size() - 1; i++) {
			if (list.get(i) > list.get(i + 1))
				return false;
		}

		return true;
	}
	
	/**
	 * Helper method to get the maximum number from a list.
	 * 
	 * @param list the ArrayList of integer elements
	 * @return the maximum number in the list
	 */
	private static int getMax(ArrayList<Integer> list) {
		int currentMax = -1;
		for (int i = 0; i < list.size(); i++) {
			if (list.get(i) > currentMax)
				currentMax = list.get(i);
		}

		return currentMax;
	}

	/**
	 * Helper method that takes the last element of a list as pivot, and places all
	 * elements smaller than pivot to the left of pivot and all greater elements to
	 * the right of pivot.
	 * 
	 * @param list the ArrayList of integer elements
	 * @param min  the beginning index of this list
	 * @param max  the ending index of this list
	 * @return index of the pivot element in this list
	 */
	private static int partition(ArrayList<Integer> list, int min, int max) {
		// Partition element (Element to be placed in the "middle")
		int pivot = list.get(max);

		// Index of smaller element
		int i = min;

		for (int j = min; j <= max - 1; j++) {
			// If current element is smaller than the pivot
			if (list.get(j) < pivot) {
				swap(list, i, j);
				i++;
			}
		}
		// Places pivot in the middle
		swap(list, i, max);
		return i;
	}

	/**
	 * Helper method that merges two sorted sub-lists into a list, resulting in a
	 * sorted list.
	 * 
	 * @param list the ArrayList of integer elements
	 * @param min  the beginning index of the first sub-list
	 * @param mid  the ending index of the first sub-list
	 * @param max  the ending index of the second sub-list
	 */
	private static void merge(ArrayList<Integer> list, int min, int mid, int max) {
		ArrayList<Integer> temp = new ArrayList<Integer>();

		int min1 = min, max1 = mid;
		int min2 = mid + 1, max2 = max;

		// Copy smaller item from each sub-list into temp until
		// one of the sub-lists is exhausted
		while (min1 <= max1 && min2 <= max2) {
			if (list.get(min1) < list.get(min2)) {
				temp.add(list.get(min1));
				min1++;
			} else {
				temp.add(list.get(min2));
				min2++;
			}
		}

		// Copy the remaining elements from first sub-list, if any
		while (min1 <= max1) {
			temp.add(list.get(min1));
			min1++;
		}

		// Copy the remaining elements from second sub-list, if any
		while (min2 <= max2) {
			temp.add(list.get(min2));
			min2++;
		}

		// Copy merged data into original list
		for (Integer val : temp) {
			list.set(min, val);
			min++;
		}

	}
}
